#app{font-family:Avenir,Helvetica,Arial,sans-serif;-webkit-font-smoothing:antialiased;-moz-osx-font-smoothing:grayscale;color:#2c3e50;margin-left:10px;margin-top:10px;margin-bottom:10px;max-width:98vw}.modal-mask{position:fixed;z-index:9998;top:0;left:0;width:100%;height:100%;background-color:rgba(0,0,0,.5);display:table;transition:opacity .3s ease}.modal-wrapper{display:table-cell;vertical-align:middle}.modal-container{width:500px;margin:0 auto;padding:20px 30px;background-color:#fff;border-radius:2px;box-shadow:0 2px 8px rgba(0,0,0,.33);transition:all .3s ease;font-family:Helvetica,Arial,sans-serif}.modal-header h3{margin-top:0;color:#42b983}.modal-body{margin:20px 0}.modal-default-button{float:right}.modal-enter,.modal-leave-active{opacity:0}.modal-enter .modal-container,.modal-leave-active .modal-container{transform:scale(1.1)}.plot-div{position:relative}.plot-close{position:absolute;z-index:1}.tab{cursor:pointer}.tab-table{max-height:83vh}.tab-table-item{padding-left:0}.var-dropdown{cursor:pointer}.fade-enter-active,.fade-leave-active{transition:opacity .5s}.fade-enter,.fade-leave-to{opacity:0}.plots-div{max-height:91vh;overflow-y:auto;border:1px blue;border-style:solid;padding-right:15px;border-radius:10px}.communication-status{position:absolute;bottom:0;right:0;height:30px;width:100px}.modal-fullscreen-container{width:90%vw;margin:0 auto;padding:20px 30px;background-color:#fff;border-radius:2px;box-shadow:0 2px 8px rgba(0,0,0,.33);transition:all .3s ease;font-family:Helvetica,Arial,sans-serif}
/*# sourceMappingURL=app.c68ba9dcf7bd08c3aad2e946c455bc61.css.map */